import { WebSocketClientConnectionProtocol, WebSocketData } from '@mswjs/interceptors/WebSocket';
import { WebSocketHandlerEventMap, WebSocketHandler } from './handlers/WebSocketHandler.mjs';
import { Path } from './utils/matching/matchRequestUrl.mjs';
import 'strict-event-emitter';

type WebSocketLink = {
    /**
     * A set of all WebSocket clients connected
     * to this link.
     */
    clients: Set<WebSocketClientConnectionProtocol>;
    on<EventType extends keyof WebSocketHandlerEventMap>(event: EventType, listener: (...args: WebSocketHandlerEventMap[EventType]) => void): WebSocketHandler;
    /**
     * Broadcasts the given data to all WebSocket clients.
     *
     * @example
     * const service = ws.link('wss://example.com')
     * service.on('connection', () => {
     *   service.broadcast('hello, everyone!')
     * })
     */
    broadcast(data: WebSocketData): void;
    /**
     * Broadcasts the given data to all WebSocket clients
     * except the ones provided in the `clients` argument.
     *
     * @example
     * const service = ws.link('wss://example.com')
     * service.on('connection', ({ client }) => {
     *   service.broadcastExcept(client, 'hi, the rest of you!')
     * })
     */
    broadcastExcept(clients: WebSocketClientConnectionProtocol | Array<WebSocketClientConnectionProtocol>, data: WebSocketData): void;
};
/**
 * Intercepts outgoing WebSocket connections to the given URL.
 *
 * @example
 * const chat = ws.link('wss://chat.example.com')
 * chat.on('connection', ({ client }) => {
 *   client.send('hello from server!')
 * })
 */
declare function createWebSocketLinkHandler(url: Path): WebSocketLink;
declare const ws: {
    link: typeof createWebSocketLinkHandler;
};

export { type WebSocketLink, ws };
