import { invariant } from "outvariant";
import { matchRequestUrl } from '../utils/matching/matchRequestUrl.mjs';
const MSW_WEBSOCKET_CLIENTS_KEY = "msw:ws:clients";
class WebSocketClientManager {
  constructor(channel, url) {
    this.channel = channel;
    this.url = url;
    this.inMemoryClients = /* @__PURE__ */ new Set();
    if (typeof localStorage !== "undefined") {
      localStorage.removeItem = new Proxy(localStorage.removeItem, {
        apply: (target, thisArg, args) => {
          const [key] = args;
          if (key === MSW_WEBSOCKET_CLIENTS_KEY) {
            this.inMemoryClients.clear();
          }
          return Reflect.apply(target, thisArg, args);
        }
      });
    }
  }
  inMemoryClients;
  /**
   * All active WebSocket client connections.
   */
  get clients() {
    if (typeof localStorage !== "undefined") {
      const inMemoryClients = Array.from(this.inMemoryClients);
      console.log("get clients()", inMemoryClients, this.getSerializedClients());
      return new Set(
        inMemoryClients.concat(
          this.getSerializedClients().filter((serializedClient) => {
            if (inMemoryClients.every(
              (client) => client.id !== serializedClient.clientId
            )) {
              return serializedClient;
            }
          }).map((serializedClient) => {
            return new WebSocketRemoteClientConnection(
              serializedClient.clientId,
              new URL(serializedClient.url),
              this.channel
            );
          })
        )
      );
    }
    return this.inMemoryClients;
  }
  getSerializedClients() {
    invariant(
      typeof localStorage !== "undefined",
      "Failed to call WebSocketClientManager#getSerializedClients() in a non-browser environment. This is likely a bug in MSW. Please, report it on GitHub: https://github.com/mswjs/msw"
    );
    const clientsJson = localStorage.getItem(MSW_WEBSOCKET_CLIENTS_KEY);
    if (!clientsJson) {
      return [];
    }
    const allClients = JSON.parse(clientsJson);
    const matchingClients = allClients.filter((client) => {
      return matchRequestUrl(new URL(client.url), this.url).matches;
    });
    return matchingClients;
  }
  addClient(client) {
    this.inMemoryClients.add(client);
    if (typeof localStorage !== "undefined") {
      const serializedClients = this.getSerializedClients();
      const nextSerializedClients = serializedClients.concat({
        clientId: client.id,
        url: client.url.href
      });
      localStorage.setItem(
        MSW_WEBSOCKET_CLIENTS_KEY,
        JSON.stringify(nextSerializedClients)
      );
    }
  }
  /**
   * Adds the given `WebSocket` client connection to the set
   * of all connections. The given connection is always the complete
   * connection object because `addConnection()` is called only
   * for the opened connections in the same runtime.
   */
  addConnection(client) {
    this.addClient(client);
    const handleExtraneousMessage = (message) => {
      const { type, payload } = message.data;
      if (typeof payload === "object" && "clientId" in payload && payload.clientId !== client.id) {
        return;
      }
      switch (type) {
        case "extraneous:send": {
          client.send(payload.data);
          break;
        }
        case "extraneous:close": {
          client.close(payload.code, payload.reason);
          break;
        }
      }
    };
    const abortController = new AbortController();
    this.channel.addEventListener("message", handleExtraneousMessage, {
      signal: abortController.signal
    });
    client.addEventListener("close", () => abortController.abort(), {
      once: true
    });
  }
}
class WebSocketRemoteClientConnection {
  constructor(id, url, channel) {
    this.id = id;
    this.url = url;
    this.channel = channel;
  }
  send(data) {
    this.channel.postMessage({
      type: "extraneous:send",
      payload: {
        clientId: this.id,
        data
      }
    });
  }
  close(code, reason) {
    this.channel.postMessage({
      type: "extraneous:close",
      payload: {
        clientId: this.id,
        code,
        reason
      }
    });
  }
}
export {
  MSW_WEBSOCKET_CLIENTS_KEY,
  WebSocketClientManager,
  WebSocketRemoteClientConnection
};
//# sourceMappingURL=WebSocketClientManager.mjs.map